/* Process model C form file: ROU_IGMP.pr.c */
/* Portions of this file Copyright (C) MIL 3, Inc. 1992 */



/* OPNET system definitions */
#include <opnet.h>
#include "ROU_IGMP.pr.h"
FSM_EXT_DECS



/* Header block */
#define TRACE if(0)
#include "ip_frag_sup.h"

#define ARRIVAL_FRM_IP		op_intrpt_type () == OPC_INTRPT_STRM 
#define UPD_INT				((op_intrpt_type () == OPC_INTRPT_SELF) && (op_intrpt_code () == 0))
#define SH_INT				((op_intrpt_type () == OPC_INTRPT_SELF) && (op_intrpt_code () == 1))

#define MAX_NO_LINKS		10
#define MAX_NO_GRPS			50

#define QOSPF_STREAM		1
#define IP_STREAM			0

#define ON					1
#define OFF					0

#define VL					if(ON)
#define IGMPDL				if(OFF)

#define IS_MULTICAST		500

#define IGMP_REFRESH_PERIOD	1

#define IGMP_PROTOCOL_ID	9
#define UDP_PROTOCOL_ID		20

typedef struct{  
	int interface[MAX_NO_LINKS];
	int mtu[MAX_NO_LINKS];
}mc_id;

typedef struct{  
	int me[MAX_NO_GRPS];
}member;



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	Objid                   		sv_my_id;
	Objid                   		sv_ip_objid;
	Objid                   		sv_qospf_objid;
	int	                    		sv_net_num0;
	int	                    		sv_node_num0;
	int	                    		sv_net_num1;
	int	                    		sv_node_num1;
	int	                    		sv_net_num2;
	int	                    		sv_node_num2;
	int	                    		sv_net_num3;
	int	                    		sv_node_num3;
	int	                    		sv_net_num4;
	int	                    		sv_node_num4;
	int	                    		sv_query_sent[100][5];
	int	                    		sv_ngrp;
	double                  		sv_upd;
	ip_dgram_list*	         		sv_igmp_list_ptr;
	int	                    		sv_grp_sent;
	int	                    		sv_n_links;
	} ROU_IGMP_state;

#define pr_state_ptr            		((ROU_IGMP_state*) SimI_Mod_State_Ptr)
#define my_id                   		pr_state_ptr->sv_my_id
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define qospf_objid             		pr_state_ptr->sv_qospf_objid
#define net_num0                		pr_state_ptr->sv_net_num0
#define node_num0               		pr_state_ptr->sv_node_num0
#define net_num1                		pr_state_ptr->sv_net_num1
#define node_num1               		pr_state_ptr->sv_node_num1
#define net_num2                		pr_state_ptr->sv_net_num2
#define node_num2               		pr_state_ptr->sv_node_num2
#define net_num3                		pr_state_ptr->sv_net_num3
#define node_num3               		pr_state_ptr->sv_node_num3
#define net_num4                		pr_state_ptr->sv_net_num4
#define node_num4               		pr_state_ptr->sv_node_num4
#define query_sent              		pr_state_ptr->sv_query_sent
#define ngrp                    		pr_state_ptr->sv_ngrp
#define upd                     		pr_state_ptr->sv_upd
#define igmp_list_ptr           		pr_state_ptr->sv_igmp_list_ptr
#define grp_sent                		pr_state_ptr->sv_grp_sent
#define n_links                 		pr_state_ptr->sv_n_links




/* Process model interrupt handling procedure */



void
ROU_IGMP ()
	{
	int 			xx,disp,j,i,proto,ip_strm,in_strm,join,term,grp_id,data_len;
	Packet			*tmp_pk1,*tmp_pk2,*igmp_pkptr,*ip_pkptr,*decap_pkptr,*pkptr;
	int			*tmp_ptr;
	mc_id			*mc_table;
	int 			SEND_RESP,dest_node,src_net,src_node,Indep_Rou;
	Objid			sub_netid;


	FSM_ENTER (ROU_IGMP)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "init")
			{
			grp_sent = 0;
			my_id = op_id_self();
			op_ima_obj_attr_get(my_id,"NGRP", &ngrp);
			
			
			ip_objid = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 0 );
			qospf_objid = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, QOSPF_STREAM );
			 
			 /* Use the object id to get the address */
			op_ima_obj_attr_get (ip_objid, "net_num0", &net_num0);
			op_ima_obj_attr_get (ip_objid, "node_num0", &node_num0);
			op_ima_obj_attr_get (ip_objid, "net_num1", &net_num1);
			op_ima_obj_attr_get (ip_objid, "node_num1", &node_num1);
			op_ima_obj_attr_get (ip_objid, "net_num2", &net_num2);
			op_ima_obj_attr_get (ip_objid, "node_num2", &node_num2);
			op_ima_obj_attr_get (ip_objid, "net_num3", &net_num3);
			op_ima_obj_attr_get (ip_objid, "node_num3", &node_num3);
			op_ima_obj_attr_get (ip_objid, "net_num4", &net_num4);
			op_ima_obj_attr_get (ip_objid, "node_num4", &node_num4);
			op_ima_obj_attr_get (ip_objid, "Indep_Rou", &Indep_Rou); 
			op_intrpt_schedule_self(op_sim_time() + 6, 0);
			/*op_intrpt_schedule_self(op_sim_time() + 5.9, 1);*/
			
			for(i = 0; i < MAX_NO_GRPS; ++i)
			{
				for(j = 0; j < 5; ++j)
					query_sent[i][j] = 0;
			}
			
			if(!Indep_Rou)
				sub_netid=op_topo_parent(op_topo_parent(ip_objid));
			else 
				sub_netid=op_topo_parent(ip_objid);
			
			n_links = op_topo_assoc_count(sub_netid,OPC_TOPO_ASSOC_IN,OPC_OBJTYPE_LKSIMP);
			
			VL	IGMPDL	printf("N-links at %d.%d = %d\n",net_num0,node_num0,n_links);
			}


		/** state (init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "init")
			{
			}


		/** state (init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle")
			{
			
			
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,ROU_IGMP)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle")
			{
			}


		/** state (idle) transition processing **/
		FSM_INIT_COND (ARRIVAL_FRM_IP)
		FSM_TEST_COND (UPD_INT)
		FSM_TEST_LOGIC ("idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 2, state2_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 3, state3_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (FRM_IP) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "FRM_IP")
			{
			in_strm = op_intrpt_strm();
			decap_pkptr = op_pk_get (in_strm);
			op_pk_nfd_get(decap_pkptr,"src_net",&src_net);
			op_pk_nfd_get(decap_pkptr,"src_node",&src_node);
			op_pk_nfd_get(decap_pkptr,"strm",&ip_strm);
			op_pk_nfd_get(decap_pkptr,"prot_id",&proto);
			op_pk_nfd_get(decap_pkptr,"data",&igmp_pkptr);
			op_pk_nfd_get(decap_pkptr,"dest_node",&dest_node);
			op_pk_nfd_get(igmp_pkptr,"terminate",&term);
			op_pk_nfd_get(igmp_pkptr,"GRP_ID",&grp_id);
			op_pk_nfd_get(igmp_pkptr,"join",&join);
			
			if(proto != IGMP_PROTOCOL_ID) 
				op_sim_end("IGMP in router is getting the wrong pks","","","");
			if(join == 1 && term == 0)
			{
				/*  int Send_Qospf_Flag = OFF;  */
				mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
				query_sent[grp_id][ip_strm] = 0;
			
				if(mc_table[grp_id].interface[ip_strm] != 1 )
					/*  Send_Qospf_Flag = ON;  */
					mc_table[grp_id].interface[ip_strm] = 1;
				op_pk_destroy(decap_pkptr);
			
				op_intrpt_schedule_remote( op_sim_time(), 6, qospf_objid );  
			
				VL IGMPDL
				{
					printf("Sending remote interrupt to QOSPF module from IGMP module\n");
					getchar();
				}
			
				op_pk_destroy(igmp_pkptr);  
			}
			
			else
			{  
				if( dest_node == IS_MULTICAST )
				{
					mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
			/*		tmp_pk1 = op_pk_copy(igmp_pkptr) ; tmp_pk2 = op_pk_copy(decap_pkptr);*/
					tmp_pk1=igmp_pkptr; tmp_pk2=decap_pkptr; /* changed by Jiemei Mar 20, 97*/
					for(j=0;j< ngrp +1; j++)  /* Group ID starts at 1 */
					{
						igmp_pkptr = op_pk_copy(tmp_pk1); 
						decap_pkptr = op_pk_copy(tmp_pk2);
						SEND_RESP = 0;
						i = 0;
						while(i < n_links + 1  && !SEND_RESP)  /*Goes to n_links+1 because interfaces go from 0 to n_links*/
					        {
			/*                                if (op_sim_time()<12)
			                                {
			  					printf("GWY_IGMP multicast:  mc_table[%3d].interface[%3d]  %3d\n",j,i,mc_table[j].interface[i]); 
			                                }    added by Jiemei for debugging Mar 21,97*/
							if(mc_table[j].interface[i] == 1 && i != ip_strm) SEND_RESP = 1;
							i++;
						}
						if(src_net == net_num0 && src_node == node_num0) SEND_RESP = 0;
						if(SEND_RESP)
						{
							op_pk_nfd_set(igmp_pkptr, "join", 1);
							op_pk_nfd_set(igmp_pkptr, "terminate", 0);
							op_pk_nfd_get(decap_pkptr, "src_net", &src_net);
							op_pk_nfd_get(decap_pkptr, "src_node", &src_node);
							op_pk_nfd_set(decap_pkptr, "dest_net", src_net);
							op_pk_nfd_set(decap_pkptr, "dest_node", src_node);
							op_pk_nfd_set(decap_pkptr, "src_net", net_num0);
							op_pk_nfd_set(decap_pkptr, "src_node", node_num0);
							op_pk_nfd_set(igmp_pkptr, "GRP_ID", j);
							data_len=op_pk_total_size_get(igmp_pkptr); /* changed by Jiemei Mar 20,97*/
							op_pk_nfd_set(decap_pkptr, "data", igmp_pkptr);
							op_pk_bulk_size_set(decap_pkptr,data_len); /*changed by Jiemei Mar 20,97*/
							op_pk_nfd_set(decap_pkptr, "ttl", 2);
							op_pk_send(decap_pkptr, IP_STREAM);
						}
						else 
						{ 
							op_pk_destroy(decap_pkptr); 
							op_pk_destroy(igmp_pkptr);
						}
					}
					op_pk_destroy(tmp_pk1);   
					op_pk_destroy(tmp_pk2);
				}
				else op_sim_end("Error in ROU_IGMP !!","","","");
			} 
			}


		/** state (FRM_IP) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "FRM_IP")
			{
			/*mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
			 
			for(i=0;i<10;++i)
			  {  j=0;
			     while(j<5)
			        { if( mc_table[i].interface[j] == 1)
			                printf("Member of grp: %d on interface %d (%d.%d)\n",i,j,net_num0,node_num0);
			          j++;
			        }
			  }
			*/
			}


		/** state (FRM_IP) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (UPD) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "UPD")
			{
			mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
			disp = 0;
			i=0;
			while(i < ngrp +1)   /* Group ID starts at 1*/
			{
				for(j=0;j < n_links+1;++j)
				{
					if(mc_table[i].interface[j] != 1)
						query_sent[i][j] == 0;
					else 
						query_sent[i][j]++;
					if(query_sent[i][j] == 2 && mc_table[i].interface[j] == 1)
					{
						disp = 1;
						IGMPDL printf("Tearing down for grp: %d on intrface %d at %d.%d\n",i,j,net_num0,node_num0);
						mc_table[i].interface[j] = 0;
					}
				}
				i++;
			}
			
			pkptr = op_pk_create_fmt("MC_APP_CNTRL");
			op_pk_nfd_set(pkptr,"join",0);
			op_pk_nfd_set(pkptr,"terminate",0);
			op_pk_nfd_set(pkptr,"GRP_ID",500);
			
			data_len = op_pk_total_size_get (pkptr) / 8;
			ip_pkptr = op_pk_create_fmt ("ip_dgram");
			op_pk_nfd_set(ip_pkptr, "data", pkptr);
			op_pk_nfd_set(ip_pkptr, "prot_id", IGMP_PROTOCOL_ID);
			op_pk_bulk_size_set(ip_pkptr, data_len * 8);
			
			op_pk_nfd_set(ip_pkptr, "dest_net", 500);
			op_pk_nfd_set(ip_pkptr, "dest_node", 500);
			
			/* Set the source address */
			op_pk_nfd_set(ip_pkptr, "src_net", net_num0);
			op_pk_nfd_set(ip_pkptr, "src_node", node_num0);
			
			/* set the length attributes of the IP packet. */
			op_pk_nfd_set(ip_pkptr, "orig_len", data_len);
			op_pk_nfd_set(ip_pkptr, "frag_len", data_len);
			op_pk_nfd_set(ip_pkptr, "ttl",2);
			
			/* Indicate that the packet is not yet fragmented. */
			op_pk_nfd_set(ip_pkptr, "frag", 0);
			op_pk_send(ip_pkptr,0); 
			
			op_intrpt_schedule_self(op_sim_time()+IGMP_REFRESH_PERIOD,0);
			
			IGMPDL if(disp)
			{
				mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
				printf("Time: %f		TEARING  IP Addr: %d.%d\n",op_sim_time(),net_num0,node_num0);
				printf("GRP   Interface   \n");
				for(i=0;i< ngrp+1;++i)
				{  
					j=0;
					while(j<5)
					{ 
						if( mc_table[i].interface[j] == 1)
						printf("%6d%12d\n",i,j);
						j++;
					}
				}
			}
			}


		/** state (UPD) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "UPD")
			{
			/*mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
			printf("%f:	GRP	Interface	IP Addr\n",op_sim_time()); 
			for(i=0;i<10;++i)
			  {  j=0;
			     while(j<5)
			        { if( mc_table[i].interface[j] == 1)
			                printf("   	%d	%d	(%d.%d)\n",i,j,net_num0,node_num0);
			          j++;
			        }
			  }
			getchar();*/
			}


		/** state (UPD) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (dis) enter executives **/
		FSM_STATE_ENTER_FORCED (4, state4_enter_exec, "dis")
			{
			/*mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
			printf("Time: %f                   JOINING Ip Addr: %d.%d\n",op_sim_time(),net_num0,node_num0);
			printf("GRP   Interface   \n");
			for(i=0;i< ngrp+1;++i)
			  {  j=0;
			     while(j<5)
			        { if( mc_table[i].interface[j] == 1)
			                printf("%6d%13d\n",i,j);
			          j++;
			        }
			  }
			getchar();
			*/
			}


		/** state (dis) exit executives **/
		FSM_STATE_EXIT_FORCED (4, state4_exit_exec, "dis")
			{
			op_intrpt_schedule_self(op_sim_time() + 60, 1);
			}


		/** state (dis) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,ROU_IGMP)
	}




void
ROU_IGMP_svar (prs_ptr,var_name,var_p_ptr)
	ROU_IGMP_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (ROU_IGMP_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("my_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_id);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("qospf_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_qospf_objid);
	if (Vos_String_Equal ("net_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num0);
	if (Vos_String_Equal ("node_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num0);
	if (Vos_String_Equal ("net_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num1);
	if (Vos_String_Equal ("node_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num1);
	if (Vos_String_Equal ("net_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num2);
	if (Vos_String_Equal ("node_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num2);
	if (Vos_String_Equal ("net_num3" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num3);
	if (Vos_String_Equal ("node_num3" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num3);
	if (Vos_String_Equal ("net_num4" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num4);
	if (Vos_String_Equal ("node_num4" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num4);
	if (Vos_String_Equal ("query_sent" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_query_sent);
	if (Vos_String_Equal ("ngrp" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ngrp);
	if (Vos_String_Equal ("upd" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_upd);
	if (Vos_String_Equal ("igmp_list_ptr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_igmp_list_ptr);
	if (Vos_String_Equal ("grp_sent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_grp_sent);
	if (Vos_String_Equal ("n_links" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_n_links);

	FOUT;
	}




void
ROU_IGMP_diag ()
	{
	int 			xx,disp,j,i,proto,ip_strm,in_strm,join,term,grp_id,data_len;
	Packet			*tmp_pk1,*tmp_pk2,*igmp_pkptr,*ip_pkptr,*decap_pkptr,*pkptr;
	int			*tmp_ptr;
	mc_id			*mc_table;
	int 			SEND_RESP,dest_node,src_net,src_node,Indep_Rou;
	Objid			sub_netid;

	FIN (ROU_IGMP_diag ())


	FOUT;
	}




void
ROU_IGMP_terminate ()
	{
	int 			xx,disp,j,i,proto,ip_strm,in_strm,join,term,grp_id,data_len;
	Packet			*tmp_pk1,*tmp_pk2,*igmp_pkptr,*ip_pkptr,*decap_pkptr,*pkptr;
	int			*tmp_ptr;
	mc_id			*mc_table;
	int 			SEND_RESP,dest_node,src_net,src_node,Indep_Rou;
	Objid			sub_netid;

	FIN (ROU_IGMP_terminate ())


	FOUT;
	}



Compcode
ROU_IGMP_init (pr_state_pptr)
	ROU_IGMP_state			**pr_state_pptr;
	{
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	FIN (ROU_IGMP_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (ROU_IGMP)",
			sizeof (ROU_IGMP_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (ROU_IGMP_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



